#include "ai.h"

#define DIST_ATTACK 6     // "Stabbing" distance between fighters (7b = 14px)
#define DIST_MINIMUM 4    // Distance for not getting too close (4b = 8px)


/**
 * Check if there's solid ground below the character when stepping backwards.
 *   O=–   –=O
 *   /\     /\
 *   =+== ==+=
 * @param dir Check left or right side of character.
 * @return 0 If ground is solid, 1 if will fall down.
 */
u8 isCloseToLedge(EDirection dir) {
   u8 tileX,
      tileY = (currChar->entity.y + G_ENEMYFIGHT_00_H) / 4;
   if (dir == DIR_LEFT) {
      tileX = currChar->entity.x / 2;
      if (tileX == 0) {
         return 0;
      }
   }
   else {
      tileX = (currChar->entity.x + G_ENEMYFIGHT_00_W) / 2 - 1;
      if (tileX >= MAP_WIDTH-1) {
         return 0;
      }
   }
   return currentScreenTiles[tileY*MAP_WIDTH+tileX] >= TILE_NO_COLLISION_INDEX;
}


/**
 * Make sure to not fall to death while following the player.
 * Checks the tiles marked with x until it reaches map height.
 * -=O
 *  /\
 *  x+–––
 *  x|
 *  x|
 */
u8 willNotFallOffScreen() {
   const u8 *mapEnd = &currentScreenTiles[0] + SCREEN_SIZE_TILES;
   u8 tileX,
      tileY = (currChar->entity.y + G_ENEMYFIGHT_00_H) / 4,
      *tilePtr;
   if (currChar->facingDirection == DIR_LEFT) {
      tileX = currChar->entity.x / 2;
      if (tileX == 0) {
         return TRUE;
      }
   }
   else {
      tileX = (currChar->entity.x + G_ENEMYFIGHT_00_W) / 2 - 1;
      if (tileX >= MAP_WIDTH-1) {
         return TRUE;
      }
   }
   tilePtr = &currentScreenTiles[tileY*MAP_WIDTH+tileX];
   while(tilePtr < mapEnd) {
      if (*tilePtr < TILE_NO_COLLISION_INDEX) {
         // Found something to stand on.
         return TRUE;
      }
      tilePtr += MAP_WIDTH;
   }
   // No solid tile found!
   return FALSE;
}


void advance() {
   if (currChar->facingDirection == DIR_LEFT) {
      currChar->nextActions = ACT_LEFT;
   }
   else {
      currChar->nextActions = ACT_RIGHT;
   }
}



void retreat() {
   if (currChar->facingDirection == DIR_LEFT) {
      currChar->nextActions = ACT_RIGHT;
   }
   else {
      currChar->nextActions = ACT_LEFT;
   }
}


void attack() {
   currChar->nextActions = ACT_UP;
}


void parry() {
   currChar->nextActions = ACT_DOWN;
}


/**
 * This function updates the actions of AI characters.
 * It uses an action cooldown to lower the frequency of performed actions.
 * It reacts to player attacks.
 * Part 1 of the other logic, executed when half of cooldown passed:
 * - don't get too close to the character
 * - make sure not to fall off ledges
 * - defend against attacks with a certain probability, slightly increasing with levels.
 * Part 2, executed once the cooldown is at 0:
 * - attack the hero with a certain probability.
 */
void updateAiMoves() {
   i8 idistX;
   u8 distX, distY;
   // Are we still sleeping?
   if (currChar->aiState == AI_IDLE || currChar->state >= ST_DYING) {
      return;
   }
   distY = faceOpponent(&hero);
   if (distY > FIGHT_MIN_DIST_Y) {
      if (willNotFallOffScreen()) {
         advance();
      }
      return;
   }
   idistX = hero.entity.x - currChar->entity.x;
   distX  = idistX >= 0 ? idistX : -idistX;
   if (distX > DIST_ATTACK) {
      if (!isCloseToLedge(currChar->facingDirection)) {
         advance();
      }
   }
   else {
      currChar->actionCooldown--;
      // Defense.
      if (   hero.state == ST_FIGHT_ATTACKING
          && currChar->state != ST_FIGHT_DEFENDING) {
         u8 decision = cpct_rand8();
         if (decision < 128 + currentLevel*5) {    // Defending gets increasingly probable over the levels.
            parry();
            currChar->actionCooldown = AI_ACTION_COOLDOWN;
         }
         if (decision < 196 + currentLevel*5) {    // Counterattacking with a slight chance.
            attack();
            currChar->actionCooldown = AI_ACTION_COOLDOWN;
         }
      }
      // Part 1.
      if (currChar->actionCooldown > AI_ACTION_COOLDOWN/2) {
         // Movement.
         EDirection side = currChar->facingDirection == DIR_LEFT ? DIR_RIGHT : DIR_LEFT;
         if (isCloseToLedge(side)) {
            advance();     // We have been pushed back towards a ledge, move away.
         }
         else if (distX <= DIST_MINIMUM) {
            retreat();     // We are too close.
         }
      }
      // Part 2.
      if (currChar->actionCooldown == 0) {         // Maybe attack.
         u8 decision = cpct_rand8();
         currChar->actionCooldown = AI_ACTION_COOLDOWN;
         if (decision < 128) {
            attack();     // Charge!
         }
      }
   }
}