#include "characters.h"

TCharacter *currChar;
u8 isCurrCharHero;         // Stores if current character is the hero.

/**
 * @param c Pointer to character struct for processing a character.
 */
inline void setCharacter(TCharacter *c) {
   currChar = c;
   isCurrCharHero = currChar == &hero;
}


/**
 * Handles most state changes. Some states are dependent on animations, so they are updated in
 * the updateAnimation() functions.
 * ## Code below could be solved better by introducing state priorities.
 * @param newState
 */
void setState(EState nextState) {
   u8 state = currChar->state;
   // We don't want to interrupt fight sequences with normal states, as well as the dying process.
   if (   nextState < FIGHT_STATES_INDEX && state >= FIGHT_STATES_INDEX
      || state >= ST_DYING) {
      return;
   }

   switch(nextState) {
      case ST_WALKING:
         if (   state != ST_JUMPING
             && state != ST_FALLING) {
            state = ST_WALKING;
         }
         break;
      case ST_JUMPING:
         if (currChar->isOnFloor) {
            state = ST_JUMPING;
            currChar->jumpIdx = 0;
         }
         break;
      case ST_IDLE:
         if (   state != ST_JUMPING
             && state != ST_FALLING) {
            state = ST_IDLE;
         }
         break;
      case ST_FIGHT_HOLDING:
         if (   state != ST_FIGHT_UNSHEATING
             && state != ST_FIGHT_SHEATING
             && state != ST_FIGHT_ATTACKING
             && state != ST_FIGHT_STEPPING
             && state != ST_FIGHT_HIT
             && state != ST_FIGHT_DEFENDING) {
            state = ST_FIGHT_HOLDING;
         }
         break;
      case ST_FIGHT_STEPPING:
         if (   state != ST_FIGHT_UNSHEATING
             && state != ST_FIGHT_SHEATING
             && state != ST_FIGHT_ATTACKING
             && state != ST_FIGHT_HIT
             && state != ST_FIGHT_DEFENDING) {
            state = ST_FIGHT_STEPPING;
         }
         break;
      case ST_FIGHT_ATTACKING:
         if (   state != ST_FIGHT_UNSHEATING
             && state != ST_FIGHT_SHEATING
             && state != ST_FIGHT_STEPPING
             && state != ST_FIGHT_HIT
             && state != ST_FIGHT_DEFENDING) {
            state = ST_FIGHT_ATTACKING;
         }
         break;
      case ST_FIGHT_DEFENDING:
         if (   state != ST_FIGHT_UNSHEATING
             && state != ST_FIGHT_SHEATING
             && state != ST_FIGHT_STEPPING
             && state != ST_FIGHT_HIT
             && state != ST_FIGHT_ATTACKING) {
            state = ST_FIGHT_DEFENDING;
         }
         break;
      default:
         state = nextState;
         break;
   }
   // Did the state change at all and is it allowed to change?
   if (currChar->state != nextState && state == nextState) {
      currChar->state = state;
      currChar->animation.frameCounter = 0;
   }
}


void cropVelocity() {
   i16 *velX = &currChar->physics.vx, *velY = &currChar->physics.vy;
   i16 maxVelX = currChar->state < FIGHT_STATES_INDEX ? MAX_VELOCITY_X : MAX_FIGHT_VELOCITY_X;
   if (*velX > maxVelX) {
      *velX = maxVelX;
   }
   else if (*velX < -maxVelX) {
      *velX = -maxVelX;
   }
   if (*velY > MAX_VELOCITY_Y) {
      *velY = MAX_VELOCITY_Y;
   }
   else if (*velX < -MAX_VELOCITY_Y) {
      *velY = -MAX_VELOCITY_Y;
   }
}


/**
 * Switch current character to fighting state.
 * AI will change to alerted once this gets called.
 * @param opponent  The other character in this fight, remains unchanged.
 */
void enterFight(TCharacter *opponent) {
   u8 offset;
   currChar->state = ST_FIGHT_UNSHEATING;
   currChar->physics.vx = 0;
   currChar->animation.redrawState = REDRAW_YES;
   // Set width and height.
   if (currChar == &hero) {
      offset = G_HEROFIGHT_00_W - G_HERO_0_W;
      currChar->animation.width = G_HEROFIGHT_00_W;
      currChar->animation.height = G_HEROFIGHT_00_H;
   }
   else {
      offset = G_ENEMYFIGHT_00_W - G_ENEMY_0_W;
      currChar->animation.width = G_ENEMYFIGHT_00_W;
      currChar->animation.height = G_ENEMYFIGHT_00_H;
      currChar->aiState = AI_ALERTED;     // Don't idle anymore from now on.
   }
   // Face opponent.
   if (currChar->entity.x < opponent->entity.x) {
      currChar->facingDirection = DIR_RIGHT;
   }
   else {
      currChar->facingDirection = DIR_LEFT;
      // Prevent "jumping" when animation with new width starts.
      setCharacterPosition(currChar, currChar->entity.x-offset, currChar->entity.y);
   }
}


/**
 * @param chr The character which stops fighting.
 */
void leaveFight(TCharacter *chr) {
   if (chr->state < ST_DYING) {
      chr->state = ST_IDLE;
   }
   chr->animation.redrawState = REDRAW_LEAVEFIGHT;
}


void takeHit();


/**
 * Checks if current character is attacked by another character and registers a hit when needed.
 * Only supposed to be called when both overlap.
 * Also stops attacker if we are defending or if we are hit.
 * @param attacker
 */
void checkAttackFrom(TCharacter *attacker) {
   const u8 attackAniDelay    = cCharacterAnimations[0][ST_FIGHT_ATTACKING].delay;
   const u8 attackAniDuration = attackAniDelay * (cCharacterAnimations[0][ST_FIGHT_ATTACKING].numFrames - 1);
   u8 stopAttacker = FALSE;
   if (currChar->state >= ST_DYING || attacker->state >= ST_DYING) {
      return;
   }
   if (attacker->state == ST_FIGHT_ATTACKING) {
      // Calculate overlapping width.
      u8 overlap;
      if (currChar->entity.x > attacker->entity.x) {
         overlap = attacker->entity.x + attacker->animation.width - currChar->entity.x;
      }
      else {
         overlap = currChar->entity.x + currChar->animation.width - attacker->entity.x;
      }
      if (   overlap >= FIGHT_HIT_OVERLAP
          && attacker->animation.frameCounter > attackAniDelay) {       // Don't score a hit when just starting to attack.
         if (currChar->state == ST_FIGHT_DEFENDING) {
            // Under attack, overlapping, attacker is "mid-swing" and we're defending.
            stopAttacker = TRUE;
            if (attacker->animation.frameCounter == attackAniDuration) {
               playSFX(SFX_DEFENDED);
            }
         }
         else if (currChar->state != ST_FIGHT_HIT) {
            // We're overlapping, not defending, not already hit and attacker is finishing his stab (animation).
            takeHit();
            if (isCurrCharHero && currChar->health == 0) {
               // Enemy killed hero.
               attacker->state = ST_FIGHT_SHEATING;
               attacker->animation.frameCounter = 0;
            }
            stopAttacker = TRUE;
         }
      }
   }
   if (stopAttacker) {
      attacker->physics.vx = 0;                          // Stop attacker.
      attacker->entity.x = attacker->entity.px;
      attacker->physics.x = attacker->entity.x * SCALE;
   }
}


/**
 * If in a fight, face opponent.
 * @param opponent
 * @return Absolute Y distance to the opponent, needed for AI.
 */
u8 faceOpponent(TCharacter *opponent) {
   i8 idistY = opponent->entity.y - currChar->entity.y;
   u8 distY  = idistY >= 0 ? idistY : -idistY;
   if (   distY < FIGHT_MIN_DIST_Y
       && currChar->state < ST_DYING) {
      if (currChar->entity.x < opponent->entity.x) {
         currChar->facingDirection = DIR_RIGHT;
      }
      else if (currChar->entity.x > opponent->entity.x) {
         currChar->facingDirection = DIR_LEFT;
      }
   }
   return distY;
}


void checkCharacterOverlaps();
void handleTileCollisions();
void handleScreenBoundary();

/**
 * Central per frame update fucntion for current character.
 * Also updates animation state (has to be done here for checking redraw over other sprites).
 */
void updateCharacter() {
   TPhysics *phys = &currChar->physics;
   TEntity *ent   = &currChar->entity;
   u8 isNotInFight = currChar->state < FIGHT_STATES_INDEX;

   ent->px = ent->x;
   ent->py = ent->y;
   currChar->walkingDirection = DIR_NONE;

   // Update the state according to character actions (inputs).
   if (currChar->nextActions == ACT_NONE) {                       // No input.
      if (isNotInFight) {
         setState(ST_IDLE);
      }
      else {
         setState(ST_FIGHT_HOLDING);
      }
   }
   else {
      if ((currChar->nextActions & ACT_LEFT) == ACT_LEFT) {
         currChar->walkingDirection = DIR_LEFT;
         if (isNotInFight) {
            currChar->facingDirection = DIR_LEFT;
            setState(ST_WALKING);
         }
         else {
            setState(ST_FIGHT_STEPPING);
         }
      }
      else if ((currChar->nextActions & ACT_RIGHT) == ACT_RIGHT) {
         currChar->walkingDirection = DIR_RIGHT;
         if (isNotInFight) {
            currChar->facingDirection = DIR_RIGHT;
            setState(ST_WALKING);
         }
         else {
            setState(ST_FIGHT_STEPPING);
         }
      }
      if ((currChar->nextActions & ACT_UP) == ACT_UP) {
         if (isNotInFight) {
            setState(ST_JUMPING);
         }
         else {
            setState(ST_FIGHT_ATTACKING);
         }
      }
      if ((currChar->nextActions & ACT_DOWN) == ACT_DOWN) {
         if (!isNotInFight) {
            setState(ST_FIGHT_DEFENDING);
         }
      }
      // Reset actions to none.
      currChar->nextActions = ACT_NONE;
   }

   // Apply gravity.
   phys->vy += GRAVITY;

   // Execute state.
   switch(currChar->state) {
      case ST_IDLE:            chrSlowDown(); break;
      case ST_WALKING:         chrWalk();     break;
      case ST_JUMPING:         chrJump();     break;
      case ST_FALLING:         chrFall();     break;
      case ST_FIGHT_HOLDING:   chrSlowDown(); break;
      case ST_FIGHT_STEPPING:  chrStep();     break;
      case ST_FIGHT_ATTACKING: chrAttack();   break;
      case ST_FIGHT_DEFENDING: chrSlowDown(); break;
      case ST_FIGHT_SHEATING:  chrSlowDown(); break;
      case ST_FIGHT_HIT:       chrHit();      break;
      case ST_DEAD:            chrDie();      break;
   }

   // Update physics.
   cropVelocity();
   phys->x += phys->vx;
   phys->y += phys->vy;
   // Calculate new screen position, move sprite if necessary.
   ent->x = phys->x / SCALE;
   ent->y = phys->y / SCALE;

   handleTileCollisions();
   handleScreenBoundary();
   checkCharacterOverlaps();

   // Did position change, i.e. no collision.
   if (ent->x != ent->px && currChar->animation.redrawState == REDRAW_NO) {
      currChar->animation.redrawState = REDRAW_YES;
   }
   if (ent->y != ent->py) {
      // Did character fall while in a fight?
      if (!isNotInFight && ent->y > ent->py) {
         leaveFight(currChar);
      }
      else if (currChar->animation.redrawState == REDRAW_NO) {
         currChar->animation.redrawState = REDRAW_YES;
      }
   }

   if (phys->vy != 0) {
      currChar->isOnFloor = FALSE;
   }
}


void chrWalk() {
   TPhysics *phys = &currChar->physics;
   switch(currChar->walkingDirection) {
      case DIR_LEFT:
         phys->vx -= ACCEL_X;
         break;
      case DIR_RIGHT:
         phys->vx += ACCEL_X;
         break;
   }
}


void chrSlowDown() {
   i16 *velX = &currChar->physics.vx;
   if (*velX > 0) {
      if (*velX <= FRICTION_X) {
         *velX = 0;
      }
      else {
         *velX -= FRICTION_X;
      }
   }
   else {
      if (*velX >= -FRICTION_X) {
         *velX = 0;
      }
      else {
         *velX += FRICTION_X;
      }
   }
}


void chrJump() {
   if (currChar->jumpIdx < JUMPSTEPS) {
      if (currChar == &hero && currChar->jumpIdx == 1) {    // A bit ugly: Don't play on the first frame.
         playSFX(SFX_JUMP);
      }
      currChar->physics.vy = -cJumpVelocities[currChar->jumpIdx];
      currChar->jumpIdx++;
   }
   else {
      currChar->jumpIdx = 0;
      setState(ST_FALLING);
   }

   if (currChar->walkingDirection == DIR_NONE) {
      chrSlowDown();
   }

   // Let character move in the air.
   chrWalk();
}


void chrFall() {
   if (currChar->walkingDirection == DIR_NONE) {
      chrSlowDown();
   }

   // Let character move in the air.
   chrWalk();
}


/**
 * Make a step in a fight.
 */
void chrStep() {
   chrWalk();
}


void chrAttack() {
   const u8 attackDelay = cCharacterAnimations[0][ST_FIGHT_ATTACKING].delay;
   if (currChar->animation.frameCounter < attackDelay) {             // Move a bit during the first frame.
      currChar->walkingDirection = currChar->facingDirection;
      chrWalk();
   }
   else {
      currChar->physics.vx = 0;
   }
}


void chrHit() {
   const u8 pushbackDelay = cCharacterAnimations[0][ST_FIGHT_HIT].delay * (cCharacterAnimations[0][ST_FIGHT_HIT].numFrames-1);
   if (currChar->animation.frameCounter < pushbackDelay) {
      return;
   }
   // Set push back direction.
   if (currChar->facingDirection == DIR_LEFT) {
      currChar->walkingDirection = DIR_RIGHT;
   }
   else {
      currChar->walkingDirection = DIR_LEFT;
   }
   chrWalk();     // Move towards push back direction.
}


void chrDie() {
   if (isCurrCharHero) {
      setGameOver(GAME_PLAYER_KILLED);
   }
   else {
      // Hero killed an enemy.
      playSFX(SFX_ENEMY_KILLED);
      // Hero killed this enemy.
      hero.state = ST_FIGHT_SHEATING;
      hero.animation.frameCounter = 0;
      currChar->entity.deletionState = DEL_QUEUED;
   }
}


void takeHit() {
   playSFX(SFX_HURT);
   currChar->health--;
   if (isCurrCharHero) {
      redrawHudFlags |= REDRAW_HUD_HEROHEALTH;
      cpct_setBorder(HW_RED);   // Dark red border.
   }
   if (currChar->health == 0) {
      if (!isCurrCharHero) {
         if (currentLevel == NUM_LEVELS-1) {
            // Special case: Final boss was killed.
            changeScore(SCORE_ENEMY_KILLED*2);
            setGameOver(GAME_LEVEL_FINISHED);
         }
         else {
            changeScore(SCORE_ENEMY_KILLED);
         }
      }
      currChar->state = ST_DYING;   // Set directly, disregard usual state logic.
      currChar->animation.frameCounter = 0;
      currChar->physics.vx = 0;
      currChar->physics.vy = 0;
   }
   else {
      setState(ST_FIGHT_HIT);
   }
}


void collectCoin(TCoin *c) __z88dk_fastcall {
   numCollectedCoins++;
   c->entity.deletionState = DEL_QUEUED;
   redrawHudFlags |= REDRAW_HUD_COINS;
   changeScore(SCORE_COIN);
   playSFX(SFX_COLLECT_COIN);
}


/**
 * Add key to inventory.
 * @param k
 */
void collectKey(TKey *k) __z88dk_fastcall {
   keyInventory |= k->color;
   k->entity.deletionState = DEL_QUEUED;
   redrawHudFlags |= REDRAW_HUD_KEYS;
   changeScore(SCORE_KEY);
   playSFX(SFX_COLLECT_KEY);
}


/**
 * Checks if the character collides with any solid tiles at its new position
 * and moves back to previous position when needed.
 * It checks for collision on the new X position only, then checks for collisions on Y axis only.
 * @param chr
 */
void handleTileCollisions() {
   TPhysics *phys  = &currChar->physics;
   TEntity *ent    = &currChar->entity;
   TAnimation *ani = &currChar->animation;
   u8 *tilePtr, tileTestX, tileTestY, tilePreviousX, tilePreviousMaxX,
      tilePreviousY, tilePreviousMaxY, *generalOffset;

   // Check for collision with tiles by moving only along Y axis.
   if (   ent->y + ani->height < (MAP_HEIGHT-1)*4-1
       && ent->y > 0) {
      tileTestY = (phys->vy < 0 ? ent->y : ent->y + ani->height - 1) / 4;
      tilePreviousX = ent->px / 2;
      tilePreviousMaxX = (ent->px + ani->width - 1) / 2;
      generalOffset = currentScreenTiles + tileTestY * MAP_WIDTH;
      for (u8 offsetX = tilePreviousX; offsetX <= tilePreviousMaxX; offsetX++) {
         tilePtr = generalOffset + offsetX;
         if (*tilePtr < TILE_NO_COLLISION_INDEX) {
            // Collide with solid tile on Y axis.
            if (phys->vy > 0) {
               currChar->isOnFloor = TRUE;
               if (currChar->state == ST_FALLING) {
                  currChar->state = ST_IDLE;
               }
            }
            else if (currChar->state == ST_JUMPING) {
               setState(ST_FALLING);
            }
            ent->y = ent->py;
            phys->y = ent->y * SCALE;
            phys->vy = 0;
            break;
         }
      }
   }
   // Check for collision with tiles by moving only along X axis.
   if (ent->x > 0 && ent->x + ani->width < (MAP_WIDTH-1)*4-1) {
      tileTestX = (phys->vx < 0 ? ent->x : ent->x + ani->width - 1) / 2;
      tilePreviousY = ent->py / 4;
      tilePreviousMaxY = (ent->py + ani->height - 1) / 4;   // Don't check lowest tile when standing on the floor.
      generalOffset = currentScreenTiles + tileTestX;
      for (u8 offsetY = tilePreviousY; offsetY <= tilePreviousMaxY; offsetY++) {
         tilePtr = generalOffset + offsetY * MAP_WIDTH;
         if (*tilePtr < TILE_NO_COLLISION_INDEX) {
            ent->x = ent->px;
            phys->x = ent->x * SCALE;
            phys->vx = 0;
            break;
         }
      }
   }
}


inline u8 isInFightingRange(TCharacter *other) {
   i16 idistY = currChar->entity.y - other->entity.y; // Can be bigger than +-127, thus two bytes.
   u8 distY = idistY >= 0 ? idistY : -idistY;
   if (distY <= FIGHT_MIN_DIST_Y) {
      i8 idistX = currChar->entity.x - other->entity.x;
      u8 distX = idistX >= 0 ? idistX : -idistX;
      if (distX <= FIGHT_MIN_DIST_X) {
         return TRUE;
      }
   }
   return FALSE;
}


/**
 * Checks overlaps between current character and other entities, setting the redraw flag as needed.
 * Also collects items etc. if current character is hero.
 * Also handles entering fights and checking for hits between characters.
 */
void checkCharacterOverlaps() {
   // Coins.
   for (i8 i = currentScreen->numCoins-1; i >= 0; i--) {
      TCoin *c = currentScreen->coins[i];
      if (   c->entity.deletionState == DEL_ACTIVE
         && isColliding(currChar->entity.x, currChar->entity.y, currChar->animation.width, currChar->animation.height,
                        c->entity.x, c->entity.y, G_COIN_W, G_COIN_H)) {
         if (isCurrCharHero) {
            collectCoin(c);
         }
         else {
            c->sprite.needsRedraw = TRUE;
            currChar->animation.redrawState = REDRAW_YES;
         }
      }
   }

   // Keys.
   for (i8 i = currentScreen->numKeys-1; i >= 0; i--) {
      TKey *k = currentScreen->keys[i];
      if (   k->entity.deletionState == DEL_ACTIVE
         && isColliding(currChar->entity.x, currChar->entity.y, currChar->animation.width, currChar->animation.height,
                        k->entity.x, k->entity.y, G_KEY_R_W, G_KEY_R_H)) {
         if (isCurrCharHero) {
            collectKey(k);
         }
         else {
            k->sprite.needsRedraw = TRUE;
            currChar->animation.redrawState = REDRAW_YES;
         }
      }
   }

   // Locks.
   for (i8 i = currentScreen->numLocks-1; i >= 0; i--) {
      TLock *l = currentScreen->locks[i];
      if (isColliding(currChar->entity.x, currChar->entity.y, currChar->animation.width, currChar->animation.height,
                     l->entity.x, l->entity.y, G_LOCK_R_W, G_LOCK_R_H)) {
         l->sprite.needsRedraw = TRUE;
         currChar->animation.redrawState = REDRAW_YES;
         // Check if player interacts and has key of the same color.
         if (   isCurrCharHero
            && cpct_isKeyPressed(definedKeys[4])
            && (keyInventory & l->color) == l->color) {
            activateLock(l);
         }
      }
   }

   // Finish.
   if (   finish.exists
       && finish.screenX == currentScreenX
       && finish.screenY == currentScreenY
       && isColliding(currChar->entity.x, currChar->entity.y, currChar->animation.width, currChar->animation.height,
                      finish.entity.x, finish.entity.y, G_FINISH_W, G_FINISH_H)) {
      finish.sprite.needsRedraw = TRUE;
      currChar->animation.redrawState = REDRAW_YES;
      if (isCurrCharHero && currChar->state < FIGHT_STATES_INDEX) {
         u8 levelComplete = TRUE;
         // All coins collected?
         for (i8 i = numCoins-1; i >= 0; i--) {
            if (allCoins[i].entity.deletionState == DEL_ACTIVE) {
               levelComplete = FALSE;
               break;
            }
         }
         if (levelComplete) {
            // Coins collected and arrived at finish.
            playSFX(SFX_LEVEL_COMPLETE);
            setGameOver(GAME_LEVEL_FINISHED);
         }
      }
   }

   // Check if hero attacks us (enemy).
   if (   !isCurrCharHero
       && isColliding(currChar->entity.x, currChar->entity.y, currChar->animation.width, currChar->animation.height,
                      hero.entity.x, hero.entity.y, hero.animation.width, hero.animation.height)) {
      checkAttackFrom(&hero);
   }

   // Check if hero is attacking enemies and overlapping.
   for (i8 i = currentScreen->numEnemies-1; i >= 0; i--) {
      TCharacter *e = currentScreen->enemies[i];
      if (e->entity.deletionState != DEL_ACTIVE) {
         continue;
      }
      if (   e != currChar
         && isColliding(currChar->entity.x, currChar->entity.y, currChar->animation.width, currChar->animation.height,
                        e->entity.x, e->entity.y, e->animation.width, e->animation.height)) {
         // Redraw the overlapping characters.
         e->animation.redrawState = REDRAW_YES;
         currChar->animation.redrawState = REDRAW_YES;
         // Check if hero is attacked by this enemy.
         if (   isCurrCharHero
             && hero.state >= FIGHT_STATES_INDEX
             && e->state < ST_DYING) {
            checkAttackFrom(e);
         }
      }
      // Check distance from hero and eventually start fighting.
      if (   e->state   < ST_DYING
          && hero.state < ST_DYING) {
         if (isCurrCharHero) {
            if (isInFightingRange(e)) {
               if (hero.isOnFloor) {
                  if (hero.state < FIGHT_STATES_INDEX) {
                     // Hero: Enter fight with enemy.
                     enterFight(e);
                  }
                  else {
                     faceOpponent(e);
                  }
               }
               if (e->isOnFloor && e->state < FIGHT_STATES_INDEX) {
                  // Enemy: Enter fight with hero.
                  currChar = e;
                  enterFight(&hero);
                  currChar = &hero;
               }
            }
         }
      }
   }
}


/**
 * Switch screen if needed, otherwise stop, die, etc.
 * The position checks work with byte wraparound.
 * ## Cut sprites at screen edges (like PoP did) and let enemies traverse screens?
 */
void handleScreenBoundary() {
   u8 screenSwitchNeeded = FALSE,
      newX = currChar->entity.x,
      newY = currChar->entity.y;
   // Left border.
   if ((currChar->entity.x == 0 || currChar->entity.x > 240) && currChar->physics.vx < 0) {
      if (isCurrCharHero && shiftToScreen(currentScreenX-1, currentScreenY)) {
         screenSwitchNeeded = TRUE;
         newX = SCREEN_WIDTH - currChar->animation.width;
      }
      else {
         newX = 0;
      }
   }
   // Right border.
   else if (currChar->entity.x >= SCREEN_WIDTH - currChar->animation.width && currChar->physics.vx > 0) {
      if (isCurrCharHero && shiftToScreen(currentScreenX+1, currentScreenY)) {
         screenSwitchNeeded = TRUE;
         newX = 0;
      }
      else {
         newX = SCREEN_WIDTH - currChar->animation.width;
      }
   }
   // Top border.
   if (currChar->entity.y > 240 && currChar->physics.vy < 0) {
      if (isCurrCharHero && shiftToScreen(currentScreenX, currentScreenY-1)) {
         screenSwitchNeeded = TRUE;
         newY = MAP_HEIGHT * 4 - currChar->animation.height;
      }
      else {
         // Stop at the top of the screen.
         newY = 0;
      }
   }
   // Bottom border.
   else if (currChar->entity.y >= MAP_HEIGHT * 4 - currChar->animation.height && currChar->physics.vy > 0) {
      if (isCurrCharHero && shiftToScreen(currentScreenX, currentScreenY+1)) {
         screenSwitchNeeded = TRUE;
         newY = 0;
      }
      else {
         newY = MAP_HEIGHT * 4 - G_HERO_0_H;
         if (isCurrCharHero) {
            setGameOver(GAME_PLAYER_FELL); // Player fell out the bottom.
         }
         else {
            // "Kill", or rather, delete this character if it fell out of the screen.
            currChar->entity.deletionState = DEL_QUEUED;
         }
      }
   }

   if (newX != currChar->entity.x) {
      currChar->entity.x = newX;
      currChar->physics.x = newX * SCALE;
      currChar->physics.vx = 0;
   }
   if (newY != currChar->entity.y) {
      currChar->entity.y = newY;
      currChar->physics.y = newY * SCALE;
      currChar->physics.vy = 0;
   }
   if (screenSwitchNeeded) {
      setCharacterPosition(currChar, newX, newY);
      // Now that we have sanitized character positions we can switch & render.
      renderScreenComplete();
   }
}


/**
 * Is called outside of normal update loop, so pass character argument.
 * @param chr
 * @param x
 * @param y
 */
void setCharacterPosition(TCharacter *chr, u8 x, u8 y) {
   chr->entity.x = chr->entity.px = x;
   chr->entity.y = chr->entity.py = y;
   chr->physics.x = x * SCALE;
   chr->physics.y = y * SCALE;
}
