#include "entities.h"

TCharacter allEnemies[MAX_ENEMIES];
TCoin allCoins[MAX_COINS];
TKey allKeys[MAX_KEYS];
TLock allLocks[MAX_LOCKS];
TBridge allBridges[MAX_BRIDGES];
TCharacter hero;
TFinish finish;

u8 numCoins, numKeys, numLocks, numBridges, numEnemies;
EKeyColor keyInventory;
u8 numCollectedCoins;


/**
 * Setup for a level.
 */
void initEntities() {
   numCoins = numKeys = numLocks = numBridges = numEnemies = numCollectedCoins = 0;
   keyInventory = COLOR_NONE;
}


/**
 * Update all entities on screen.
 */
void updateEntities() {
   setCharacter(&hero);
   updateCharacter();
   for (i8 i = currentScreen->numEnemies-1; i >= 0; i--) {
      setCharacter(currentScreen->enemies[i]);
      if (currChar->entity.deletionState != DEL_ACTIVE) {
         // Deleted character, carry on.
         continue;
      }
      updateAiMoves();
      updateCharacter();
   }
}


void prepareBridgeOffscreen(TScreen *scr, EDirection dir, u8 tileY) {
   u8 alreadyExists = FALSE,
      tileX = dir == DIR_RIGHT ? 0 : MAP_WIDTH - 1;
   for (i8 i = scr->numBridges-1; i >= 0; i--) {
      TBridge *b = scr->bridges[i];
      if (b->direction == dir && b->tileY == tileY && b->tileX == tileX) {
         // We already have a bridge here, it will be extended once its screen is displayed.
         b->isExtended = TRUE;
         alreadyExists = TRUE;
      }
   }
   if (!alreadyExists) {
      // Create a new bridge.
      spawnBridge(tileX, tileY, scr, dir);
      scr->bridges[scr->numBridges-1]->isExtended = TRUE;
   }
}


void placeBridgeTile(u8 posX, u8 posY) {
   cpct_etm_drawTileBox2x4(posX, posY-1, 1, 2, MAP_WIDTH, MAP_POS, currentScreenTiles);
   playSFX(SFX_BRIDGE_TILE);
   cpct_waitHalts(30);
}


void extendBridge(TBridge *b, u8 isOnScreen) {
   u8 bridgeX = b->tileX, posX, posY = b->tileY, *tile, goesRight = b->direction == DIR_RIGHT;
   if (isOnScreen) {
      if (b->isExtended) {
         return;
      }
      b->isExtended = TRUE;
   }

   if (bridgeX > 0 && goesRight) {
      bridgeX++;
   }
   else if (bridgeX < MAP_WIDTH-1 && !goesRight){
      bridgeX--;
   }
   tile = &currentScreenTiles[posY * MAP_WIDTH + bridgeX];
   posX = bridgeX;
   if (b->direction == DIR_RIGHT) {
      while (posX < MAP_WIDTH && *tile > TILE_NO_COLLISION_INDEX - 1) {
         *tile             = posX % 2 ? TILE_BRIDGE_LOWER_1 : TILE_BRIDGE_LOWER_2;   // Lower (solid) tiles.
         *(tile-MAP_WIDTH) = posX % 2 ? TILE_BRIDGE_UPPER_1 : TILE_BRIDGE_UPPER_2;   // Upper (cosmetic) tiles.
         if (isOnScreen) {
            placeBridgeTile(posX, posY);
         }
         tile++;
         posX++;
      }
      if (posX == MAP_WIDTH) {
         TScreen *screenPtr = currentScreen + 1;
         // Reached right edge of the screen.
         if (currentScreenX < MAX_LEVEL_WIDTH - 1 && screenPtr) {
            prepareBridgeOffscreen(screenPtr, DIR_RIGHT, posY);
         }
      }
      else {
         *tile = TILE_ENTITY_BRIDGE_L;   // Solid block reached, place end of bridge.
         if (isOnScreen) {
            placeBridgeTile(posX, posY);
         }
      }
   }
   else {
      while (posX != 255 && *tile > TILE_NO_COLLISION_INDEX - 1) {
         *tile             = posX % 2 ? TILE_BRIDGE_LOWER_1 : TILE_BRIDGE_LOWER_2;   // Lower (solid) tiles.
         *(tile-MAP_WIDTH) = posX % 2 ? TILE_BRIDGE_UPPER_1 : TILE_BRIDGE_UPPER_2;   // Upper (cosmetic) tiles.
         if (isOnScreen) {
            placeBridgeTile(posX, posY);
         }
         tile--;
         posX--;
      }
      if (posX == 255) {        // If posX was 0, the last decrement caused a wraparound to 255.
         // Reached left edge of the screen.
         TScreen *screenPtr = currentScreen - 1;
         if (currentScreenX > 0 && screenPtr) {
            prepareBridgeOffscreen(screenPtr, DIR_LEFT, posY);
         }
         posX = 0; // Needed for correct rendering.
      }
      else {
         *tile = TILE_ENTITY_BRIDGE_R;   // Solid block reached, place end of bridge.
         if (isOnScreen) {
            placeBridgeTile(posX, posY);
         }
      }
   }
}


void activateLock(TLock *l) {
   u8 minDist = calcManhattanDistance(0, 0, 80, 160);   // Screen diagonal.
   TBridge *closestBridge = 0;
   keyInventory &= ~l->color; // Remove key from inventory.
   redrawHudFlags |= REDRAW_HUD_KEYS;
   // Activate closest bridge.
   for (u8 i = 0; i < currentScreen->numBridges; i++) {
      TBridge *b = currentScreen->bridges[i];
      u8 distance = calcManhattanDistance(l->entity.x, l->entity.y, b->tileX*2, b->tileY*4);
      if (distance <= minDist) {
         minDist = distance;
         closestBridge = b;
      }
   }
   if (closestBridge) {
      extendBridge(closestBridge, TRUE);
      changeScore(SCORE_BRIDGE_EXT);
   }
}


void initPhysics(TCharacter *chr) __z88dk_fastcall {
   TPhysics *phys = &chr->physics;
   phys->x = chr->entity.x * SCALE;
   phys->y = chr->entity.y * SCALE;
   phys->vx = 0;
   phys->vy = 0;
}


void initEntity(TEntity *ent, u8 posX, u8 posY) {
   ent->x = posX;
   ent->y = posY;
   ent->px = posX;
   ent->py = posY;
   ent->deletionState = DEL_ACTIVE;
}


/**
 * Initializes everything except health and AI states because these differ with hero vs. enemies.
 * @param c
 * @param posX
 * @param posY
 */
void initCharacter(TCharacter *c, u8 posX, u8 posY) {
   c->state = ST_IDLE;
   initEntity(&c->entity, posX, posY);
   initPhysics(c);
   initCharacterAnimation(c);
   c->actionCooldown = AI_ACTION_COOLDOWN;
   c->aiState = AI_IDLE;
   c->facingDirection = DIR_LEFT;
   c->isOnFloor = FALSE;
   c->jumpIdx = 0;
   c->nextActions = ACT_NONE;
   c->walkingDirection = DIR_NONE;
}

void spawnCoin(u8 posX, u8 posY) {
   TCoin *newCoin = &allCoins[numCoins];
   initEntity(&newCoin->entity, posX, posY);
   newCoin->sprite.dataPtr = g_coin;
   newCoin->sprite.width = G_COIN_W;
   newCoin->sprite.height = G_COIN_H;
   addCoin(currentScreen, newCoin);
   numCoins++;
}

void spawnKey(u8 posX, u8 posY, EKeyColor color) {
   TKey *newKey = &allKeys[numKeys];
   initEntity(&newKey->entity, posX, posY);
   newKey->color = color;
   switch (color) {
      case COLOR_RED:
         newKey->sprite.dataPtr = g_key_r;
         break;
      case COLOR_GREEN:
         newKey->sprite.dataPtr = g_key_g;
         break;
      case COLOR_BLUE:
         newKey->sprite.dataPtr = g_key_b;
         break;
      case COLOR_YELLOW:
         newKey->sprite.dataPtr = g_key_y;
         break;
   }
   newKey->sprite.width = G_KEY_R_W;
   newKey->sprite.height = G_KEY_R_H;
   addKey(currentScreen, newKey);
   numKeys++;
}

void spawnLock(u8 posX, u8 posY, EKeyColor color) {
   TLock *newLock = &allLocks[numLocks];
   initEntity(&newLock->entity, posX, posY);
   newLock->color = color;
   switch (color) {
      case COLOR_RED:
         newLock->sprite.dataPtr = g_lock_r;
         break;
      case COLOR_GREEN:
         newLock->sprite.dataPtr = g_lock_g;
         break;
      case COLOR_BLUE:
         newLock->sprite.dataPtr = g_lock_b;
         break;
      case COLOR_YELLOW:
         newLock->sprite.dataPtr = g_lock_y;
         break;
   }
   newLock->sprite.width = G_LOCK_R_W;
   newLock->sprite.height = G_LOCK_R_H;
   addLock(currentScreen, newLock);
   numLocks++;
}


void spawnBridge(u8 tileX, u8 tileY, TScreen *scr, EDirection dir) {
   TBridge *newBridge = &allBridges[numBridges];
   newBridge->tileX = tileX;
   newBridge->tileY = tileY;
   newBridge->isExtended = FALSE;
   newBridge->direction = dir;
   addBridge(scr, newBridge);
   numBridges++;
}


void spawnFinish(u8 posX, u8 posY) {
   initEntity(&finish.entity, posX, posY);
   finish.exists = TRUE;
   finish.screenX = currentScreenX;
   finish.screenY = currentScreenY;
   finish.sprite.dataPtr = g_finish;
   finish.sprite.width = G_FINISH_W;
   finish.sprite.height = G_FINISH_H;
}


void spawnPlayer(u8 posX, u8 posY) {
   initCharacter(&hero, posX, posY);
   hero.health = HEALTH_HERO_START;
}

void spawnEnemy(u8 posX, u8 posY) {
   TCharacter *e = &allEnemies[numEnemies];
   initCharacter(e, posX, posY);
   e->health = cEnemyHealthLevels[currentLevel];
   addEnemy(currentScreen, e);
   numEnemies++;
}

