#ifndef CONSTANTS_H
#define CONSTANTS_H

#include <types.h>
#include <video/colours.h>

///////////////////////////////////////////////////////////////////////////////
// Constant definitions.
//

#define TRUE 1
#define FALSE 0
#define SCREEN_WIDTH 80
#define SCREEN_HEIGHT 200
#define NUM_LEVELS 6
#define PATTERN_BLACK     (u16) (cpctm_px2byteM0(1, 1) << 8) + cpctm_px2byteM0(1, 1)      // 2 byte pen 1  pattern.
#define PATTERN_RED       (u16) (cpctm_px2byteM0(5, 5) << 8) + cpctm_px2byteM0(5, 5)      // 2 byte pen 3  pattern.
#define PATTERN_DARKBLUE  (u16) (cpctm_px2byteM0(2, 2) << 8) + cpctm_px2byteM0(2, 2)      // 2 byte pen 13 pattern.
#define PATTERN_WHITE     (u16) (cpctm_px2byteM0(15, 15) << 8) + cpctm_px2byteM0(15, 15)  // 2 byte pen 15 pattern.
#define PEN_ENEMY_LIGHT 6    // Palette index of light enemy color.
extern const u8 const cEnemyColorValues[NUM_LEVELS];

// Map constants.
#define MAP_WIDTH 40
#define MAP_HEIGHT 40
#define MAP_POS_Y SCREEN_HEIGHT-160
#define MAX_LEVEL_WIDTH 3
#define MAX_LEVEL_HEIGHT 3
#define MAX_SCREENS MAX_LEVEL_WIDTH * MAX_LEVEL_HEIGHT
#define SCREEN_SIZE_TILES MAP_WIDTH * MAP_HEIGHT
#define MAP_POS cpctm_screenPtr(CPCT_VMEM_START, 0, MAP_POS_Y)

// Special indices in the tileset.
#define TILE_NO_COLLISION_INDEX 19   // Tiles with a lower index are solid.
#define TILE_ENTITY_INDEX    40
#define TILE_ENTITY_KEY_R    TILE_ENTITY_INDEX
#define TILE_ENTITY_KEY_B    TILE_ENTITY_INDEX+1
#define TILE_ENTITY_KEY_G    TILE_ENTITY_INDEX+2
#define TILE_ENTITY_KEY_Y    TILE_ENTITY_INDEX+3
#define TILE_ENTITY_LOCK_R   TILE_ENTITY_INDEX+4
#define TILE_ENTITY_LOCK_B   TILE_ENTITY_INDEX+5
#define TILE_ENTITY_LOCK_G   TILE_ENTITY_INDEX+6
#define TILE_ENTITY_LOCK_Y   TILE_ENTITY_INDEX+7
#define TILE_ENTITY_PLAYER   TILE_ENTITY_INDEX+8
#define TILE_ENTITY_ENEMY    TILE_ENTITY_INDEX+9
#define TILE_ENTITY_COIN     TILE_ENTITY_INDEX+10
#define TILE_ENTITY_FINISH   TILE_ENTITY_INDEX+11
#define TILE_ENTITY_BRIDGE_L 5
#define TILE_ENTITY_BRIDGE_R 6
#define TILE_BRIDGE_UPPER_1  32     // Upper 1212
#define TILE_BRIDGE_UPPER_2  33     //       ----
#define TILE_BRIDGE_LOWER_1  13     // Lower ====
#define TILE_BRIDGE_LOWER_2  14     //       1212

// Maximum entities in the game.
#define MAX_ENEMIES 5
#define MAX_COINS 32
#define MAX_KEYS 4
#define MAX_LOCKS 4
#define MAX_BRIDGES 4

// Maximum entities per screen.
#define SCREEN_MAX_ENEMIES 2
#define SCREEN_MAX_COINS 5
#define SCREEN_MAX_KEYS 2
#define SCREEN_MAX_LOCKS 2
#define SCREEN_MAX_BRIDGES 2

// Animations.
#define FRAMEDELAY_SHORT 9
#define FRAMEDELAY_MEDIUM 18
#define FRAMEDELAY_LONG 30

// Character constants.
#define HEALTH_HERO_START  4
#define SCORE_COIN         10
#define SCORE_KEY          20
#define SCORE_ENEMY_KILLED 100
#define SCORE_BRIDGE_EXT   20
#define SCORE_LEVEL_DONE   200
#define FIGHT_MIN_DIST_Y   3*4      // Min. Y distance of three tiles to start a fight.
#define FIGHT_MIN_DIST_X   15*2     // Min. X distance.
#define FIGHT_HIT_OVERLAP  2        // How much two characters have to overlap in order to hit each other.
#define AI_ACTION_COOLDOWN 60
extern const u8 const cEnemyHealthLevels[NUM_LEVELS];

// Physics constants (all in Subpixel units)
#define SCALE 256    // Scale of the physics world (in subpixels).
#define ACCEL_X 12
#define FRICTION_X 16
#define MAX_VELOCITY_X 90
#define MAX_FIGHT_VELOCITY_X 40
#define MAX_VELOCITY_Y 786
#define GRAVITY 50
#define JUMPSTEPS 15
extern const u16 const cJumpVelocities[JUMPSTEPS];

///////////////////////////////////////////////////////////////////////////////
// Enums.
// Many of these are used as bit fields.
//

// HUD redraw flags.
typedef enum {
   REDRAW_HUD_NOTHING = 0,
   REDRAW_HUD_COINS = 1,
   REDRAW_HUD_KEYS = 2,
   REDRAW_HUD_SCORE = 4,
   REDRAW_HUD_HEROHEALTH = 8,
   REDRAW_HUD_ENEMYHEALTH = 16,
} ERedrawFlag;

// Character actions.
typedef enum {
   ACT_NONE = 0,
   ACT_LEFT = 1,
   ACT_RIGHT = 2,
   ACT_UP = 4,
   ACT_DOWN = 8,
   ACT_INTERACT = 16, // e.g. pick up key, use key on lock, etc.
} EAction;

// Facing direction.
typedef enum {
   DIR_LEFT,
   DIR_RIGHT,
   DIR_NONE,
} EDirection;


// Character redraw mode.
typedef enum {
   REDRAW_NO,
   REDRAW_YES,
   REDRAW_LEAVEFIGHT,
} ECharRedraw;


// Color of locks and keys.
typedef enum {
   COLOR_NONE = 0,
   COLOR_RED = 1,
   COLOR_GREEN = 2,
   COLOR_BLUE = 4,
   COLOR_YELLOW = 8,
} EKeyColor;

// Character states.
typedef enum {
   ST_IDLE,
   ST_WALKING,
   ST_JUMPING,
   ST_FALLING,
   ST_INTERACTING,
   ST_FIGHT_UNSHEATING,   // Character is drawing sword.
   ST_FIGHT_ATTACKING,
   ST_FIGHT_HOLDING,
   ST_FIGHT_STEPPING,
   ST_FIGHT_DEFENDING,
   ST_FIGHT_SHEATING,     // Character is putting back sword.
   ST_FIGHT_HIT,
   ST_DYING,
   ST_DEAD,
} EState;
#define FIGHT_STATES_INDEX ST_FIGHT_UNSHEATING  // Character states >= this mark that it's in a fight.
#define NUM_CHAR_STATES ST_DEAD+1

// State for AI logic.
typedef enum {
   AI_IDLE,          // Doing nothing, no fight yet.
   AI_ALERTED,       // Always following player.
} EAiState;

// Flags for entity deletion state.
typedef enum {
   DEL_ACTIVE,
   DEL_QUEUED,
   DEL_INACTIVE
} EDeletionState;

// Game state.
typedef enum {
   GAME_RUNNING,
   GAME_PLAYER_KILLED,
   GAME_PLAYER_FELL,
   GAME_RESTART_LEVEL,
   GAME_BACK_TO_MENU,
   GAME_LEVEL_FINISHED
} EGameState;

///////////////////////////////////////////////////////////////////////////////
// Data structures.
//

typedef struct {
   EDirection direction;
   u8 *dataPtr;
} TFrame;

typedef struct {
   u8 *dataPtr;
   u8 width, height, needsRedraw;
} TSprite;

typedef struct {
   TFrame *pFrame;
   u8 width, height, frameCounter;
   ECharRedraw redrawState;
} TAnimation;

typedef struct {
   u16 x, y;
   i16 vx, vy;
} TPhysics;

typedef struct {
   u8 x, y;
   u8 px, py;   // Previous position.
   EDeletionState deletionState;
} TEntity;

typedef struct {
   TEntity entity;
   TPhysics    physics;
   TAnimation  animation;
   u8 jumpIdx, isOnFloor, health;
   u8 actionCooldown;         // Used for AI actions.
   EDirection walkingDirection, facingDirection;
   EState     state;
   EAiState   aiState;
   EAction    nextActions;
} TCharacter;

typedef struct {
   TEntity entity;
   TSprite sprite;
} TCoin;

typedef struct {
   TEntity entity;
   TSprite sprite;
   EKeyColor color;
} TKey;

typedef struct {
   TEntity entity;
   TSprite sprite;
   EKeyColor color;
} TLock;

typedef struct {
   u8 tileX, tileY;
   u8 isExtended;
   EDirection direction;
} TBridge;

typedef struct {
   TEntity entity;
   TSprite sprite;
   u8 screenX, screenY, exists;
} TFinish;

typedef struct {
   u8 levelX, levelY;
   TCharacter *enemies[SCREEN_MAX_ENEMIES];
   TCoin      *coins[SCREEN_MAX_COINS];
   TKey       *keys[SCREEN_MAX_KEYS];
   TLock      *locks[SCREEN_MAX_LOCKS];
   TBridge    *bridges[SCREEN_MAX_BRIDGES];
   u8 numCoins, numKeys, numLocks, numBridges, numEnemies;
} TScreen;

#endif /* CONSTANTS_H */
