#include "levels.h"

u8 currentScreenTiles[SCREEN_SIZE_TILES];
u8 currentScreenX;
u8 currentScreenY;


/**
 * Handle switching away from current screen.
 */
void changeScreenFrom() {
   if (hero.state >= FIGHT_STATES_INDEX) {
      leaveFight(&hero);
   }
   for (i8 i = currentScreen->numEnemies-1; i >= 0; i--) {
      TCharacter *e = currentScreen->enemies[i];
      if (e->entity.deletionState == DEL_ACTIVE && e->state >= FIGHT_STATES_INDEX) {
         // ## Move to safe offset from screen edge to allow player to reenter later.
         leaveFight(e);
      }
   }
}


/**
 * Handle entering new (current) screen.
 */
void changeScreenTo() {
   // Place bridge tiles as needed in freshly decrunched screen.
   for (i8 i = currentScreen->numBridges-1; i >= 0; i--) {
      TBridge *b = currentScreen->bridges[i];
      if (b->isExtended) {
         extendBridge(b, FALSE);
      }
   }
}

/**
 * Decrunches a screen of the current level and returns TRUE if successful, FALSE if no screen is present.
 * @param scrX
 * @param scrY
 */
u8 decrunchScreen(u8 scrX, u8 scrY) {
   void *decrunched_data_end, *crunched_data_end;
   u8 offset = scrY * MAX_LEVEL_WIDTH + scrX;
   if (*(allCrunchedScreens[currentLevel] + offset) == 0) {
      // No screen at this position.
      return FALSE;
   }
   // Pointer to start of crunched screen data + crunched data size - 1.
   crunched_data_end = (*(allCrunchedScreens[currentLevel] + offset)) + (*(allCrunchedScreenSizes[currentLevel] + offset)) - 1;
   decrunched_data_end = currentScreenTiles + SCREEN_SIZE_TILES - 1;
   cpct_zx7b_decrunch_s(decrunched_data_end, crunched_data_end);
   return TRUE;
}


void calculatePositionOnScreen(u16 tileIndex, u8 *tileX, u8 *tileY) {
   *tileX = tileIndex % MAP_WIDTH * 2;   // Four pixels / two bytes width.
   *tileY = tileIndex / MAP_WIDTH * 4;   // Four pixels height.
}


/**
 * Parses the tiles in a screen and spawns entities if isInitial is true.
 * @param scrX
 * @param scrY
 * @param isInitial
 * @param playerStartX
 * @param playerStartY
 */
void parseScreen(u8 scrX, u8 scrY, u8 isInitial, u8 *playerStartX, u8 *playerStartY) {
   u8 posX, posY = 0;
   if (isInitial) {
      // Initialize new screen.
      currentScreen             = &allScreens[scrY * MAX_LEVEL_WIDTH + scrX];
      currentScreen->levelX     = scrX;
      currentScreen->levelY     = scrY;
      currentScreen->numBridges = 0;
      currentScreen->numCoins   = 0;
      currentScreen->numEnemies = 0;
      currentScreen->numKeys    = 0;
      currentScreen->numLocks   = 0;
      currentScreenX = scrX;
      currentScreenY = scrY;
   }
   // Loop over all tiles in this screen and spawn entities if necessary.
   for (i16 tileIdx = SCREEN_SIZE_TILES; tileIdx > 0; tileIdx--) {
      switch(currentScreenTiles[tileIdx]) {
         case TILE_ENTITY_COIN  :
            if (isInitial) {
               calculatePositionOnScreen(tileIdx, &posX, &posY);
               spawnCoin(posX, posY);
            }
            currentScreenTiles[tileIdx] = currentScreenTiles[tileIdx+MAP_WIDTH];
            break;
         case TILE_ENTITY_KEY_R :
            if (isInitial) {
               calculatePositionOnScreen(tileIdx, &posX, &posY);
               spawnKey(posX, posY, COLOR_RED);
            }
            currentScreenTiles[tileIdx] = currentScreenTiles[tileIdx+MAP_WIDTH];
            break;
         case TILE_ENTITY_KEY_G :
            if (isInitial) {
               calculatePositionOnScreen(tileIdx, &posX, &posY);
               spawnKey(posX, posY, COLOR_GREEN);
            }
            currentScreenTiles[tileIdx] = currentScreenTiles[tileIdx+MAP_WIDTH];
            break;
         case TILE_ENTITY_KEY_B :
            if (isInitial) {
               calculatePositionOnScreen(tileIdx, &posX, &posY);
               spawnKey(posX, posY, COLOR_BLUE);
            }
            currentScreenTiles[tileIdx] = currentScreenTiles[tileIdx+MAP_WIDTH];
            break;
         case TILE_ENTITY_KEY_Y :
            if (isInitial) {
               calculatePositionOnScreen(tileIdx, &posX, &posY);
               spawnKey(posX, posY, COLOR_YELLOW);
            }
            currentScreenTiles[tileIdx] = currentScreenTiles[tileIdx+MAP_WIDTH];
            break;
         case TILE_ENTITY_LOCK_R:
            if (isInitial) {
               calculatePositionOnScreen(tileIdx, &posX, &posY);
               spawnLock(posX, posY, COLOR_RED);
            }
            currentScreenTiles[tileIdx] = currentScreenTiles[tileIdx+MAP_WIDTH];
            break;
         case TILE_ENTITY_LOCK_G:
            if (isInitial) {
               calculatePositionOnScreen(tileIdx, &posX, &posY);
               spawnLock(posX, posY, COLOR_GREEN);
            }
            currentScreenTiles[tileIdx] = currentScreenTiles[tileIdx+MAP_WIDTH];
            break;
         case TILE_ENTITY_LOCK_B:
            if (isInitial) {
               calculatePositionOnScreen(tileIdx, &posX, &posY);
               spawnLock(posX, posY, COLOR_BLUE);
            }
            currentScreenTiles[tileIdx] = currentScreenTiles[tileIdx+MAP_WIDTH];
            break;
         case TILE_ENTITY_LOCK_Y:
            if (isInitial) {
               calculatePositionOnScreen(tileIdx, &posX, &posY);
               spawnLock(posX, posY, COLOR_YELLOW);
            }
            currentScreenTiles[tileIdx] = currentScreenTiles[tileIdx+MAP_WIDTH];
            break;
         case TILE_ENTITY_PLAYER:
            if (isInitial) {
               calculatePositionOnScreen(tileIdx, &posX, &posY);
               spawnPlayer(posX, posY);
               *playerStartX = scrX;
               *playerStartY = scrY;
            }
            currentScreenTiles[tileIdx] = currentScreenTiles[tileIdx+MAP_WIDTH];
            break;
         case TILE_ENTITY_ENEMY :
            if (isInitial) {
               calculatePositionOnScreen(tileIdx, &posX, &posY);
               spawnEnemy(posX, posY);
            }
            currentScreenTiles[tileIdx] = currentScreenTiles[tileIdx+MAP_WIDTH];
            break;
         case TILE_ENTITY_FINISH:
            if (isInitial) {
               calculatePositionOnScreen(tileIdx, &posX, &posY);
               spawnFinish(posX, posY);
            }
            currentScreenTiles[tileIdx] = currentScreenTiles[tileIdx+MAP_WIDTH];
            break;
         case TILE_ENTITY_BRIDGE_L:
            if (isInitial) {
               calculatePositionOnScreen(tileIdx, &posX, &posY);
               spawnBridge(posX/2, posY/4, currentScreen, DIR_LEFT);
            }
            break;
         case TILE_ENTITY_BRIDGE_R:
            if (isInitial) {
               calculatePositionOnScreen(tileIdx, &posX, &posY);
               spawnBridge(posX/2, posY/4, currentScreen, DIR_RIGHT);
            }
            break;
      }
   }
}


/**
 * Decrunches all screen in the level consecutively and spawns entities.
 * @param num
 */
void decrunchAndPrepareLevel() {
   u8 playerScreenX = 0, playerScreenY = 0;
   // Loop over all screens in this level.
   for (u8 scrY = 0; scrY < MAX_LEVEL_HEIGHT; scrY++) {
      for (u8 scrX = 0; scrX < MAX_LEVEL_WIDTH; scrX++) {
         if (decrunchScreen(scrX, scrY)) {
            parseScreen(scrX, scrY, TRUE, &playerScreenX, &playerScreenY);
         }
      }
   }

   if (currentLevel == 0) {
      // Set up bridge to castle (saves memory and is easier).
      prepareBridgeOffscreen(&allScreens[playerScreenY * MAX_LEVEL_WIDTH + playerScreenX], DIR_RIGHT, 27);
   }
   // Change palette to alter enemy appearance.
   cpct_setPALColour(PEN_ENEMY_LIGHT, cEnemyColorValues[currentLevel]);
   // Start level at screen with player spawn.
   shiftToScreen(playerScreenX, playerScreenY);
}


/**
 * Decrunches all tilemap screens consecutively and scans for entity tiles, then sets up
 * entities for every screen.
 * @param num
 */
void initLevel() {
   finish.exists = FALSE;
   decrunchAndPrepareLevel();
}


/**
 * @param newX
 * @param newY
 * @return TRUE if no screen is located there, FALSE if it worked.
 */
u8 shiftToScreen(u8 newX, u8 newY) {
   // Check if new position is wrapped around or out of bounds.
   if (newX == 255 || newY == 255 || newX == MAX_LEVEL_WIDTH || newY == MAX_LEVEL_HEIGHT) {
      return FALSE;
   }
   // Decrunch new screen and remove placeholders tiles.
   if (decrunchScreen(newX, newY)) {
      u8 offset = newY * MAX_LEVEL_WIDTH + newX;
      changeScreenFrom();
      currentScreenX = newX;
      currentScreenY = newY;
      currentScreen = &allScreens[offset];
      parseScreen(currentScreenX, currentScreenY, FALSE, 0, 0);
      changeScreenTo();
      return TRUE;
   }
   else {
      return FALSE;
   }
}